import datetime
import json

import dateutil

utctz = datetime.timezone.utc

class ClockAlarm(object):
    """ Represent an alarm """

    def __init__(self, _utcnow=lambda: datetime.datetime.now(tz=utctz)):
        """ Instanciate a new alarm
            @param _utcnow : a function returning current UTC datetime
        """
        self.__utcnow = _utcnow
        self._time = datetime.time(0, 0, 0, tzinfo=datetime.timezone.utc)
        self._ringing = False
        self._next_ring = None
        self._snooze = False

    @property
    def is_on(self) -> bool:
        """ @return True if the alarm is on (will ring) else False """
        return self._next_ring is not None

    @property
    def json(self) -> str:
        """ @return a json representation of the instance """
        ret = {'time':self._time.isoformat(), 'ringing':self._ringing,
               'snooze':self._snooze}
        if self._next_ring:
            ret['next_ring']  = self._next_ring.isoformat()
        return json.dumps(ret)

    def _from_json(self, data:str):
        """ Load internal state from json (as given by ClockAlarm.json property)
        """
        data = json.loads(data)
        self._time = datetime.time.fromisoformat(data['time'])
        self._ringing = data['ringing']
        self._next_ring = None if 'next_ring' not in data else data['next_ring']
        if self._next_ring:
            self._next_ring = datetime.datetime.fromisoformat(self._next_ring)
        self._snooze = data['snooze']
        return self

    def get_time(self, tz:datetime.tzinfo=utctz):
        """ @return The tz aware alarm ringing datetime.time """
        return self._get_next_ring().astimezone(tz=tz).timetz()

    def set_time(self, ringtime:datetime.time, tz:datetime.tzinfo=utctz) -> None:
        """ Set the ringtime
            @param ringtime : non-tz aware datetime.time
            @param tz : the given ringtime timezone
        """
        ringtime_dt = self.get_next_gte_datetime(ringtime, tz)
        self._time = ringtime_dt.astimezone(tz=utctz).timetz()
        if self.is_on:
            self.__set_next_ring()

    def set_state(self, on:bool) -> None:
        """ Set the ON/OFF state
            @param on : if True the alarm will ring at configured time
        """
        if on:
            self.__set_next_ring()
        else:
            self._next_ring = None
            self._ringing = False
            self._snooze = False

    def snooze(self, minutes:int=15) -> bool:
        """ Stop ringing alarm for specified amount of minutes
            @param minutes : amount of time after wich the alarm should ring again
            @return False if alarm was not ringing else True
        """
        if not self._ringing:
            return False
        self._next_ring = self.__utcnow() + datetime.timedelta(minutes=minutes)
        self._ringing = False
        self._snooze = True
        return True

    def is_ringing(self) -> bool:
        """ Check if an alarm should be ringing
            @return True if the alarm is ringing else False
        """
        if not self.is_on:
            return False
        if self._ringing:
            return True
        utcnow = self.__utcnow()
        if self._next_ring < utcnow:
            self._ringing = True
            self._snooze = False
            self.__set_next_ring()

        return self._ringing

    def is_snooze(self) -> bool:
        """ @return True if an alarm is snoozing (ringing again in a specified
                amount of time after a ClockAlarm.snooze() call)
        """
        return self._snooze

    def _get_next_ring(self) -> datetime.datetime:
        """ @return The next ringing time in utc """
        return self.get_next_gte_datetime(self._time)

    def __set_next_ring(self) -> None:
        """ Set the next ring time of the alarm """
        self._next_ring=self._get_next_ring()


    @staticmethod
    def get_next_gte_datetime(wanted:datetime.time,
            tz:datetime.tzinfo=utctz) -> datetime.datetime:
        """ Return the next datetime with a specific time
            @param wanted : the wanted time, timezone info will be dropped
            @param tz : the wanted datetime timezone
            @return The next datetime > now with wanted time
        """
        tznow_orig = datetime.datetime.now(tz=tz).replace(microsecond=0)
        tznow = tznow_orig.replace(hour=tznow_orig.hour)
        while True:
            wanted_dt = tznow.replace(hour=wanted.hour,
                                      minute=wanted.minute,
                                      second=wanted.second)
            if wanted_dt > tznow_orig:
                break
            tznow += datetime.timedelta(days=1)
        return wanted_dt
