import argparse
import datetime
import json

from .errors import AlarmNameError
from .clock import Clock
from .command import Command


def tzset_cmd(clk:Clock, args:argparse.Namespace) -> str:
    """ Function called when "tzset" command received
        @param clk : the Clock instance
        @param args : the arguments parsed by argparse
        @return a well formatted response ready to be sent through the websocket
    """
    if clk.set_tz(args.tzname):
        return Command.fmt_ok()
    return Command.fmt_err('Invalid timezone %r' % args.tzname)


# tzset command declaration
tzset = Command('tzset', tzset_cmd, description='Set the clock timezone')
tzset.add_argument('tzname', help='The new timezone (UTC for utc timezone)')


def alarm_cmd(clk:Clock, args:argparse.Namespace) -> str:
    """ Function called when "alarm *" commmand received
        @param clk : the Clock instance
        @param args : the arguments parsed by argparse
        @return a well formatted response ready to be sent through the websocket
        @throws NotImplementedError if an alarm subcommand was added but not
                implemented
    """
    cmd = args.subcmd
    ret_details = ''
    if hasattr(args, 'name') and args.name is not None and cmd != 'add':
        # try to fetch alarm if needed
        try:
            alrm = None if args.name is None else clk.alarm_get(args.name)
        except AlarmNameError as expt:
            return Command.fmt_err(expt)

    try:
        if cmd == 'add':
            clk.alarm_add(args.name)
        elif cmd == 'del':
            clk.alarm_del(args.name)
        elif cmd == 'list':
            if args.all:
                ret = list(clk.alarms.keys())
            else:
                ret = clk.alarm_rings()

            alarms = [(name, clk.alarms[name]) for name in ret]

            ret = {name:{'time':alrm.get_time(clk.tz).strftime('%H:%M:%S'),
                         'on': alrm.is_on,
                         'ringing': alrm.is_ringing(),
                         'snooze': alrm.is_snooze()}
                   for name, alrm in alarms}
            ret_details = json.dumps(ret)
        elif cmd == 'snooze':
            alrm.snooze(args.minutes)
        elif cmd == 'set':
            ringtime = datetime.time(args.hour, args.minute, args.second)
            alrm.set_time(ringtime, clk.tz)
        elif cmd == 'on':
            alrm.set_state(True)
        elif cmd == 'off':
            alrm.set_state(False)
        else:
            raise NotImplementedError('Command %r not implemented' % cmd)
    except AlarmNameError as expt:
        return Command.fmt_err(expt)

    return Command.fmt_ok(ret_details)


# parser template for alarm name
_name_parser =  argparse.ArgumentParser(add_help=False)
_name_parser.add_argument('name', type=str, help='The alarm name')

# alarm command & sub-commands declarations
alrm = Command('alarm', alarm_cmd, description='Alarm handling')

subalrm = alrm.add_subparsers(required=True, dest='subcmd')

subalrm.add_parser('add', help='Add an alarm', parents=[_name_parser])
subalrm.add_parser('del', help='Delete an alarm', parents=[_name_parser])
subalrm.add_parser('on', help='Set an alarm ON', parents=[_name_parser])
subalrm.add_parser('off', help='Set an alarm OFF', parents=[_name_parser])

alrm_snooze = subalrm.add_parser('snooze', help='Snooze ringing alarms',
                                 parents=[_name_parser])
alrm_snooze.add_argument('minutes', nargs='?', default=15, type=float,
                         help='Ring again after this amount of minutes')

alrm_set = subalrm.add_parser('set', help='Set an alarm ringing time',
                              parents=[_name_parser])
alrm_set.add_argument('hour', metavar='HH', type=int, choices=range(24))
alrm_set.add_argument('minute', metavar='MM', type=int, choices=range(60),
                      default=0, nargs='?')
alrm_set.add_argument('second', metavar='SS', type=int, choices=range(60),
                      default=0, nargs='?')

alrm_list = subalrm.add_parser('list', help='List ringing alarms returning a \
JSON object')
alrm_list.add_argument('-a', '--all', action='store_true', default=False,
                       help='List all alarms')
