import asyncio
import logging
import time

import websockets
from websockets.server import WebSocketServerProtocol

from .clock import Clock
from .command import Command

logger = logging.getLogger(__name__)


async def websocket_handler(sock_cli:WebSocketServerProtocol,
                            path:str=None) -> None:
    """ Called for each connecting client
        @param sock_cli : Client socket
        @param path : (depends on websockets version?) websocket path
    """
    cli_addr = sock_cli.remote_address[0]
    if 'X-Real-IP' in sock_cli.request_headers:
        ## TODO add a configuration to indicate expected IP from
        #  reverse proxy to check if the connection comes from it.
        cli_addr = sock_cli.request_headers['X-Real-IP']
    r_addr = '%s(%d)' % (cli_addr,
                         sock_cli.remote_address[1])
    logger.info('%s connected' % r_addr)

    # Clock restore/creation given a session_id
    try:
        session_id = await sock_cli.recv()

        clock = Clock(session_id=session_id) # per socket Clock  instance

        await sock_cli.send('SESSION:'+clock.session_id)

        last_sent = 0 # last time we sent the hour to the client (timestamp)
        last_tz = None


        while sock_cli.open:
            # Send date/time + alarms if not sent for more than 1s
            if time.time() - last_sent >= 1:
                if last_tz != clock.tzname:
                    last_tz = clock.tzname
                    await sock_cli.send('TZN:%s' % last_tz)
                await sock_cli.send(clock.now().isoformat())
                last_sent = time.time()
                for alrm in clock.alarm_rings():
                    await sock_cli.send('ALRM:%s' % alrm)
            # Attempt to read message from client
            tleft = time.time() - last_sent + 1
            tleft = .01 if tleft < .01 else tleft
            try:
                message = await asyncio.wait_for(sock_cli.recv(), tleft)
                logger.debug('%s << %r' % (r_addr, message))
                ret = Command.run_command(clock, message)
                logger.debug('%s >> %r' % (r_addr, ret))
                await sock_cli.send(ret)
            except asyncio.TimeoutError:
                pass # no message in socket
    except websockets.exceptions.ConnectionClosedOK:
        logger.info('%s closed connection' %  r_addr)
    try:
        del clock
    except UnboundLocalError:
        pass
