import logging
import json
import os
import os.path
import pathlib
import secrets
import tempfile
import time
import random

from .clock_alarm import ClockAlarm

SESSION_EXPIRE = 3600*24
MAX_SESSIONS = 0x1000

TOKEN_SZ = 16

logger = logging.getLogger(__name__)

class SessionHandler(object):
    """ Abstract class for session handler definition """

    def __init__(self, session_id:str):
        """ Instanciate a new session with a valid session_id """
        self.__content = dict()
        self.__session_id = self.validate_session_id(session_id)
        if self.__class__.__name__ == 'SessionHandler':
            raise NotImplementedError()

    def tick(self):
        """ Change expiration time"""
        raise NotImplementedError()

    def save(self):
        """ Save the session """
        return None

    @property
    def session_id(self) -> str:
        """ Return the session's ID """
        return self.__session_id

    @property
    def content(self) -> dict:
        """ Return the session's dict """
        return self.__content # ! no protection against write by reference !

    def __getitem__(self, key:str):
        """ Allows [] access """
        return self.__content[key]

    def __setitem__(self, key:str, value:any):
        """ Allows [] access """
        self.__content[key] = value

    def __delitem__(self, key:str):
        """ Delete given item from session """
        del self.__content[key]

    def __contains__(self, key:str):
        """ Return true if key is present is session's data """
        return key in self.__content

    def __iter__(self):
        """ Iterate over session data keys """
        return iter(self.__content.keys())

    @classmethod
    def _purge(cls):
        """ Purge expired sessions """
        raise NotImplementedError()

    @classmethod
    def validate_session_id(cls, session_id:str) -> str:
        """ If given session_id is valid return it, else return a new valid
            session id """
        if not len(session_id) == TOKEN_SZ*2:
            return cls.new_id()
        try:
            int(session_id, 16)
        except Exception:
            return cls.new_id()
        return session_id

    @classmethod
    def new_id(cls) -> str:
        """ Return a new session id """
        return secrets.token_hex(TOKEN_SZ)


class RamSession(SessionHandler):
    """ Session handler storing sessions in memory (all data are lost
        when the server's process is stopped """

    _sessions = dict()
    _expires = dict()

    def __init__(self, session_id:str):
        """ Create/Restor a session """
        session_id = self.validate_session_id(session_id)
        if session_id in self.__class__._sessions:
            logger.info('Restoring session %r' % session_id)
            content = self.__class__._sessions[session_id]
        else:
            session_id = self.new_id()
            logger.info('New session %r' % session_id)
            content = dict()
        super().__init__(session_id)

        for key,val in content.items():
            self[key] = val

        self.tick()
        self._purge() # purge after ticking, allowing late session restore

    def tick(self):
        """ Update session's expiration time """
        self._expires[self.session_id] = time.time() + SESSION_EXPIRE

    def __del__(self):
        """ Save session data & update expiration time """
        self._sessions[self.session_id] = self.content
        self.tick()

    @classmethod
    def _purge(cls):
        """ Delete expired sessions """
        min_expire = None
        min_key = None
        for key in cls._sessions:
            # Delete expired sessions
            if key not in cls._expires or cls._expires[key] <= time.time():
                del cls._sessions[key]
                if key in cls._expires:
                    del cls._expires[key]
            # Looks for older session in case we reached the MAX_SESSIONS count
            if min_expire is None or cls._expires[key] < min_expire:
                min_expire = cls._expires[key]
                min_key = key

        if len(cls._sessions) >= MAX_SESSIONS - 1:
            if min_key is None:
                random.choice(list(cls._sessions.keys()))
            del cls._sessions[min_key]
            if min_key in self._expires:
                del cls._expires[min_key]


class FileSessionJsonEncoder(json.JSONEncoder):
    """ Json encoder able to encode ClockAlarm instances """

    def default(self, obj):
        """ Default encoder """
        if isinstance(obj, ClockAlarm):
            return obj.json
        return json.JSONEncoder.default(self, obj)

class FileSession(SessionHandler):
    """ Session handler storing sessions into files """

    _session_dir = os.path.join(tempfile.gettempdir(), 'pyws_clock_sessions')

    def __init__(self, session_id:str):
        """ Create/Restore a session """
        if not os.path.isdir(self.__class__._session_dir):
            os.mkdir(self.__class__._session_dir)
        self._purge()

        session_id = self.validate_session_id(session_id)

        self.__path = pathlib.Path(self.getpath(session_id))
        if not os.path.isfile(self.path):
            logger.info('New session %s' % self.path)
            content = dict()
            session_id = self.new_id()
            self.__path = pathlib.Path(self.getpath(session_id))
            self.path.touch()
        else:
            logger.info('Restoring session %s' % self.path)
            with self.path.open('r') as fp:
                try:
                    content = json.load(fp)
                    content['alarms'] = {k:ClockAlarm()._from_json(v)
                                         for k,v in content['alarms'].items()}
                    if not isinstance(content, dict):
                        raise TypeError('Corrupted session file %s' % self.path)
                except (TypeError,json.decoder.JSONDecodeError) as expt:
                    logger.error(expt)
                    content = dict()
        super().__init__(session_id)

        for k,v in content.items():
            self[k] = v

        self.save()


    @property
    def path(self) -> pathlib.Path:
        """ Return session's file path """
        return self.__path

    def save(self):
        """ Save session to disk """
        with self.path.open('w+') as fp:
            json.dump(self.content, fp, cls=FileSessionJsonEncoder)

    def __del__(self):
        """ Save session to disk """
        logger.debug('Saving session in %s' % self.__path)
        self.save()

    def tick(self):
        """ Update file session modification time """
        self.path.touch()

    @classmethod
    def _purge(cls):
        """ Delete expired sessions """
        min_expire = None
        min_path = None
        session_count = 0
        for fname in os.listdir(cls._session_dir):
            session_count += 1
            path = os.path.join(cls._session_dir, fname)
            mtime = os.path.getmtime(path)
            if mtime + SESSION_EXPIRE <= time.time():
                os.remove(path)
            # Looks for older session in case we reached the MAX_SESSIONS count
            if min_expire is None or mtime < min_expire:
                min_expire = mtime
                min_path = path

        if session_count >= MAX_SESSIONS - 1:
            if min_path is not None:
                os.remove(min_path)

    @classmethod
    def getpath(cls, name:str):
        """ Given a session id return a file path """
        return os.path.join(cls._session_dir, name)

    @classmethod
    def new_id(cls):
        """ Create a new uniq session id """
        cls._purge()
        for _ in range(1024):
            res = super().new_id()
            if not os.path.isfile(cls.getpath(res)):
                return res
        err = 'Unable to create new session is %r' % cls._session_dir
        raise RuntimeError(err)


class Session(object):
    """ Wrapper for FileSession and RamSession handlers """

    _handler = RamSession

    def __init__(self):
        raise NotImplementedError()

    @classmethod
    def get(cls, session_id:str):
        """ Create/Restore a session given its ID """
        return cls._handler(session_id)

    @classmethod
    def set_file_handler(cls, tempdir:str=None):
        """ Configure the session handle to FileSession """
        if tempdir is not None:
            FileSession._session_dir = tempdir
        cls._handler = FileSession
